/*
 * main.c
 *
 * Created: 30.01.2017 21:58:39
 * Author : Chris
 */ 

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/boot.h>

#include "uart.h"
#include "simpleio.h"
#include "mx8p.h"
#include "prozeda.h"
#include "common.h"
#include "config.h"
#include "statusled.h"
#include "fifo.h"
#include "adc.h"
#include "sysinfo.h"

uint8_t proz_logdata[PROZEDA_MSGLEN_LOGDATA];
uint8_t proz_displaydata[PROZEDA_MSGLEN_DISPLAY];
uint8_t proz_headerdata[PROZEDA_MSGLEN_HEADER];
uint8_t headercnt = PIROZEDA_HEADERCNT;

typedef enum 
{
	cmd_eol_none = 0,
	cmd_eol_normal = 1,
	cmd_eol_bootloader = 2,
}	cmd_eol_t;

volatile FIFO16_t infifo;
volatile cmd_eol_t cmd_eol = cmd_eol_none;

// interrupts

ISR(SPI_STC_vect)
{
	prozeda_rxData(SPDR);
}

ISR(TIMER0_COMPA_vect)
{
	prozeda_tick();
}

ISR(USART_RX_vect)
{
	char c = UDR0;
	if(c == '\n')
	{
		cmd_eol = cmd_eol_normal;
	}
	else
	{
		FIFO16_write(infifo, c);
		if(c == BOOTLOADER_PASSWORD[3])
		{
			cmd_eol = cmd_eol_bootloader;
		}
	}
}

// boot infos

inline uint16_t adc2voltage(uint16_t adcval)
{
	uint32_t tmp = PIROZEDA_SUPPLY_AVGCNT * ADC_REF_1V1 * ADC_MAX * 1000;
	return tmp / adcval;
}

void sysvoltage_print(void)
{
	uint16_t tmp = adc_read(adc_ref_AVcc, adc_mux_1V1, PIROZEDA_SUPPLY_AVGCNT);
	uart_writeln("Info: Supply voltage = %u mV (%u)", adc2voltage(tmp), tmp);
}

void bootinfo(bool afterdark)
{
	uart_writeln("Built on: " __DATE__ " " __TIME__);
	uart_writeln("Version: " PIROZEDA_VERSION);
	
	system_resetinfo(afterdark);
	
	// fuse info and config check
	
	uint8_t fuse_e = boot_lock_fuse_bits_get(GET_EXTENDED_FUSE_BITS);
	uint8_t fuse_h = boot_lock_fuse_bits_get(GET_HIGH_FUSE_BITS);
	uint8_t fuse_l = boot_lock_fuse_bits_get(GET_LOW_FUSE_BITS);
	uart_writeln("Info: Fuses = 0x%2x%2x%2x (EHL)", fuse_e, fuse_h, fuse_l);
	
	if(fuse_e != 0xFC)
		uart_writeln("Info: FuseE should be 0xFC if bootloader is used");
	if(fuse_h != 0xC5)
		uart_writeln("Info: FuseH should be 0xC5");
	if(fuse_l != 0xDE)
		uart_writeln("Info: FuseL should be 0xDE");
		
	sysvoltage_print();
	
	uart_writeln("Info: Bootloader password is " BOOTLOADER_PASSWORD);
		
	uart_writeln("Info: Watchdog timeout = " PIROZEDA_WDT_TIMEOUT_TEXT);
	
	// enabled message types
	uart_write("Messages: LOG");

# ifdef PROZEDA_SUPPORT_DISPLAY
	uart_write(" DISP");
#endif

#ifdef PROZEDA_SUPPORT_HEADER
	uart_write(" HEAD");
#endif

#ifdef PROZEDA_MSGTYPE_DEBUG
	uart_write(" MSGTYPE");
#endif

#ifdef PROZEDA_MSGCHK_DEBUG
	uart_write(" MSGCHK_DBG");
#endif

#ifdef PIROZEDA_KITTY_DEBUG
	uart_write(" KITTY_DBG");
#endif

#ifdef PIROZEDA_TIMESTAMPS
	uart_write(" TIMESTAMPS");
#endif
	uart_writeln();
}

#ifdef PROZEDA_STATS
void stats_print(void)
{
	uart_write("Stats: logdata: %u", prozeda_data.stats_msg_logdata);
	uart_write(" display: %u", prozeda_data.stats_msg_display);
	uart_write(" header: %u", prozeda_data.stats_msg_header);
	uart_write(" rxerr_len: %u", prozeda_data.stats_rxerr_len);
	uart_writeln(" rxerr_checksum: %u", prozeda_data.stats_rxerr_checksum);
}
#else
#define stats_print()
#endif

// timestamps

#ifdef PIROZEDA_TIMESTAMPS
uint32_t timestamp = 0;

inline void timestamp_tick(void)
{
	timestamp++;
}

void timestamp_print(void)
{
	uart_write("%9u0\t", timestamp);
}
#else
#define timestamp_tick()
#define timestamp_print()
#endif

// uptime 

#ifdef PIROZEDA_UPTIME
uint8_t uptime_postscaler = 0;
inline void uptime_tick(void)
{
	uptime_postscaler++;
	if(uptime_postscaler > 100)
	{
		uptime_postscaler = 0;
		system_uptime_tick();
	}
}

#else
#define uptime_tick()
#endif

// timer configuration

inline void timer0_init(void)
{
	// setup timer
	TCCR0A = timer0a_wgm_ctc;
	TCCR0B = timer0b_wgm_ctc | timer0_prescaler_8;
	
	// ticks - this is used for the prozeda decoder
	// OCR0A = (time * F_CPU) / prescaler
	// pre 8 | 100 s => 150
	// pre 8 |  75 s => 112.5
	// pre 8 |  50 s =>  75
	OCR0A = 112;
	TIMSK0 = 1<<OCIE0A;
}

inline void timer1_init(void)
{
	// shall tick at 2.5ms
	// F_CPU / clkdiv * time = 12M / 8 * 2.5m = 3750
	OCR1A = 3750;
	TCCR1A = timer1a_coma_normal | timer1a_comb_normal | timer1a_wgm_4_CTCOcr1a;
	TCCR1B = timer1b_wgm_4_CTCOcr1a | timer1b_clkdiv_8;
	TIMSK1 = 0; //no interrupts
}

uint8_t timer1_postscaler = 0;
inline void timer1_task(void)
{
	if(TIFR1 & (1<<OCF1A))
	{
		TIFR1 |= (1<<OCF1A);
		statusled_tick();
		
		// postscale to 10 ms
		timer1_postscaler++;
		if(timer1_postscaler > 4)
		{
			timer1_postscaler = 0;
			timestamp_tick();
			uptime_tick();	
		}
	}
}

// command processor



inline bool fwupcheck(void)
{	
	// check characters arrived 
	// last character doesn't need to be checked because it's the 
	// entry condition for this method
	if(FIFO16_peek_w(infifo, 3) != BOOTLOADER_PASSWORD[0]) { return false; }
	if(FIFO16_peek_w(infifo, 2) != BOOTLOADER_PASSWORD[1]) { return false; }
	if(FIFO16_peek_w(infifo, 1) != BOOTLOADER_PASSWORD[2]) { return false; }
	
	uart_writeln("Info: Go to Bootloader");
	reset();
	return true;
}

inline void cmd_task(void)
{
	char fifolin[17];
	if(cmd_eol == cmd_eol_normal)
	{
		FIFO16_gets(infifo, fifolin);
		cmd_eol = cmd_eol_none;
		FIFO_init(infifo);
		
		if(arrcmp(fifolin, "reset", 5))
		{
			uart_writeln("Cmd: Reset");
			reset();
		}
#ifdef PIROZEDA_UPTIME
		else if(arrcmp(fifolin, "uptime", 6))
		{
			uart_writeln("Info: Uptime %u s", system_uptime_get());
		}
#endif
		else if(arrcmp(fifolin, "bootinfo", 8))
		{
			bootinfo(false);
		}
		else if(arrcmp(fifolin, "header", 6))
		{
			uart_writeln("Info: Enabled headers for %u entries", PIROZEDA_HEADERCNT);
			headercnt = PIROZEDA_HEADERCNT;
		}
#ifdef PROZEDA_STATS
		else if(arrcmp(fifolin, "stats", 5))
		{
			stats_print();
		}
#endif
		else
		{
			uart_writeln("Cmd '%s' not supported", fifolin);
		}
	}
	else if(cmd_eol == cmd_eol_bootloader)
	{
		// the bootloader entry is a little odd command-wise. 
		// at least the last character for the bootloader was received, 
		// the other ones need to be checked
		if(fwupcheck() == false)
		{
			cmd_eol = cmd_eol_none;
		}	
	}
}


int main(void)
{
	statusled_init();
	
	wdt_enable(PIROZEDA_WDT_TIMEOUT);
	
	uart_init();
	
	// init input UART
	FIFO_init(infifo);
	UCSR0B |= (1<<RXCIE0) | (1<<RXEN0);
	
	// init SPI as slave
	IOPIN_O(SPI_MISO);
	IOPIN_I(SPI_CS);
	IOPIN_H(SPI_CS); //enable pull-up for chip select
	
	SPCR = (1<<SPIE) | (1<<SPE) | (0<<DORD) | (0<<MSTR) | (0<<CPOL) | (0<<CPHA);
	
	// timer0 is for prozeda related stuff
	timer0_init();
	
	// timer1 is for general purpose (LED, timestamps)
	timer1_init();
	
	
	//init debug pins
	//IOPIN_O(DBG0);
	//IOPIN_L(DBG0);
	//IOPIN_O(DBG1);
	//IOPIN_L(DBG1);
	//IOPIN_O(DBG2);
	//IOPIN_L(DBG2);

	// assign pointers for prozeda
	prozeda_data.logdata_ptr = (uint8_t *)&proz_logdata;
	prozeda_data.displaydata_ptr = (uint8_t *)&proz_displaydata;
	prozeda_data.headerdata_ptr = (uint8_t *)&proz_headerdata;
		
	sei();
	//maximum count of header lines to be output
	
	bootinfo(true);

#ifdef PIROZEDA_UPTIME
	system_uptime_reset();
#endif

    while (1)
    {
		
		timer1_task();
		cmd_task();
		
		prozeda_taskstatus_t status = prozeda_task();

		
		if(status == prozeda_taskstatus_rxLogdata)
		{
			timestamp_print();
			uart_write("Logdata:");
			uart_rawprint(proz_logdata, PROZEDA_MSGLEN_LOGDATA);
			uart_writeln();
			wdt_reset();
			statusled_rxok();
		}
		else if(status == prozeda_taskstatus_rxDisplaydata)
		{
			timestamp_print();
			uart_write("Display:");
			uart_rawprint(proz_displaydata, PROZEDA_MSGLEN_DISPLAY);
			uart_writeln();
			wdt_reset();
			statusled_rxok();	
		}
		else if(status == prozeda_taskstatus_rxHeaderdata)
		{
			if(headercnt > 0)
			{
				timestamp_print();
				uart_write("Header:");
				uart_rawprint(proz_headerdata, PROZEDA_MSGLEN_HEADER);
				uart_write("\r\n");
				wdt_reset();	
				headercnt--;
				if(headercnt == 0)
				{
					uart_writeln("Info: Disabled headers");
				}
				
				statusled_rxok();
			}
		}
		else if(status == prozeda_taskstatus_rxErrorLen)
		{
			timestamp_print();
#ifdef PROZEDA_MSGCHK_DEBUG
			uart_writeln("rxErr:Len e:%2x, a:%2x", prozeda_data.msglen_exp, prozeda_data.msglen_act);
#else
			uart_writeln("rxErr:Len");
#endif
			statusled_rxerr();
		}
		else if(status == prozeda_taskstatus_rxErrorChecksum)
		{
			timestamp_print();
#ifdef PROZEDA_MSGCHK_DEBUG
			uart_writeln("rxErr:CRC e:%2x, a:%2x", prozeda_data.checksum_exp, prozeda_data.checksum_act);
#else
			uart_writeln("rxErr:CRC");
#endif
			statusled_rxerr();
		}
#ifdef PROZEDA_MSGTYPE_DEBUG
		else if(status == prozeda_taskstatus_rxMsgtype)
		{
			timestamp_print();
			uart_writeln("MsgType:%4x", prozeda_data._msgmode);			
		}
#endif
#ifdef PIROZEDA_KITTY_DEBUG
		else if(status == prozeda_taskstatus_kitty)
		{
			timestamp_print();
			uart_writeln("Info: Meow");
		}
#endif
    }
}